import sys, os
import signal
import trio
import inspect
from trio import lowlevel as trio_ll
from contextlib import contextmanager, asynccontextmanager, ExitStack, AsyncExitStack

async def aenumerate(aiter, start=0):
    idx = start
    async for elem in aiter:
        yield (idx, elem)
        idx += 1

class LineReader:
    def __init__(self, stream):
        self.stream = stream
        self._line_generator = self._generate_lines()
        self.last_line = None

    async def _generate_lines(self):
        buf = bytearray()
        find_start = 0
        while True:
            newline_idx = buf.find(b'\n', find_start)
            if newline_idx < 0:
                more_data = await self.stream.receive_some(1024)
                find_start = len(buf)

                if not more_data:
                    return
                buf += more_data

            else:
                line = buf[:newline_idx+1]
                del buf[:newline_idx+1]
                find_start = 0
                self.last_line = line.decode('utf-8').rstrip()
                yield self.last_line

    def __aiter__(self):
        return self._line_generator

    async def readline(self):
        try:
            return await self._line_generator.__anext__()
        except StopAsyncIteration:
            return None

@asynccontextmanager
async def signal_cancellable(signals=[signal.SIGINT, signal.SIGTERM]):
    """Allows a block of trio code to be cancelled with a signal.
    Typically wrapped around your main function to allow clean shutdown of your program."""
    async with trio.open_nursery() as nursery:
        async def signal_watcher(*, task_status=trio.TASK_STATUS_IGNORED):
            with trio.open_signal_receiver(*signals) as signal_aiter:
                task_status.started()
                async for signum in signal_aiter:
                    print(f"Caught signal {signum}, quitting")
                    nursery.cancel_scope.cancel()
                    return
        await nursery.start(signal_watcher)
        yield

@asynccontextmanager
async def with_task(func, *args, soon=False, cancel=True):
    """Async context manager that opens a nursery and runs a task in it. Especially useful with AsyncExitStack."""

    async with trio.open_nursery() as nursery:
        if soon:
            nursery.start_soon(func, *args)
        else:
            await nursery.start(func, *args)
        try:
            yield
        finally:
            if cancel:
                nursery.cancel_scope.cancel()


_stdin = None
async def trio_input():
    global _stdin
    if _stdin is None:
        _stdin = LineReader(trio_ll.FdStream(os.dup(0)))
    return await _stdin.readline()

async def receive_exactly(stream, size):
    r = []
    total = 0
    while total < size:
        chunk = await stream.receive_some(size)
        if not chunk: raise EOFError
        r.append(chunk)
        total += len(chunk)
    return b''.join(r)

class Tasklet:
    """Wraps a bit of code, allows cancelling it and joining (waiting for it to really finish).
    Like trio.CancelScope, this object should not be reused. I.e., it should be used only once in a `with' statement."""
    def __init__(self):
        self.cancel_scope = trio.CancelScope()
        self.finished_event = trio.Event()

    def __enter__(self):
        self.cancel_scope.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.finished_event.set()
        return self.cancel_scope.__exit__(exc_type, exc_value, traceback)

    async def join(self):
        return await self.finished_event.wait()

    async def cancel_wait(self, *, grace_period=0):
        if grace_period:
            with trio.move_on_after(grace_period):
                await self.join()
                return
        self.cancel_scope.cancel()
        await self.join()

    def cancel_nowait(self):
        self.cancel_scope.cancel()


@asynccontextmanager
async def open_joinable_nursery():
    ev = trio.Event()
    try:
        async with trio.open_nursery() as nursery:
            nursery.join_event = ev
            nursery.join = ev.wait
            async def cancel_wait():
                nursery.cancel_scope.cancel()
                await event.wait()
            nursery.cancel_wait = cancel_wait
            yield nursery
    finally:
        ev.set()

class _WatchInstrument:
    def __init__(self):
        self._watches = {}
        self._last_id = 0

    async def wait_true(self, predicate):
        """Wait for a predicate (parameter-less lambda) to become True.

        The predicate must be a synchronous function that is fast to evaluate and without side-effects.
        It works by checking the predicate each time before returning to the main loop."""
        if predicate(): return # already satisfied
        src = inspect.getsource(predicate).strip()
        self._last_id += 1
        id = self._last_id
        ev = trio.Event()
        if not self._watches:
            self._install()
        self._watches[id] = wd = dict(type='predicate', predicate=predicate, event=trio.Event())
        #print("Waiting for", src)
        try:
            while True:
                await wd['event'].wait()
                if 'exception' in wd: raise wd['exception']
                if predicate(): return
                wd['event'] = trio.Event() # it may have already became False again before control got to us
        finally:
            del self._watches[id]
            if not self._watches: self._uninstall()
            #print("Done waiting", src)

    def after_task_step(self, task):
        for w in self._watches.values():
            if w['type'] == 'predicate' and 'exception' not in w:
                try:
                    r = w['predicate']()
                    if r: w['event'].set()
                except Exception as exc:
                    w['exception'] = exc
                    w['event'].set()

    def _install(self):
        trio_ll.add_instrument(self)

    def _uninstall(self):
        trio_ll.remove_instrument(self)

_watch_instrument = _WatchInstrument()

wait_true = _watch_instrument.wait_true

@asynccontextmanager
async def cancel_when(predicate):
    """
    A context manager that cancels its body when a predicate (parameterless lambda) becomes true.
    """
    if predicate():
        # If predicate is already true, cancel as soon as possible (at first checkpoint in body)
        with trio.CancelScope() as cs:
            cs.cancel()
            yield
    else:
        async with trio.open_nursery() as nursery:
            async def stop_task():
                await wait_true(predicate)
                #print("cancel_when cancelling")
                nursery.cancel_scope.cancel()
            nursery.start_soon(stop_task)
            try:
                yield
            finally:
                nursery.cancel_scope.cancel()



